/*
    Copyright 2009 Luigi Auriemma

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA

    http://www.gnu.org/licenses/gpl-2.0.txt
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <sys/stat.h>

#ifdef WIN32
    #include <windows.h>
    #include <wincrypt.h>
#else
    #define stricmp strcasecmp
    #define stristr strcasestr
#endif

typedef unsigned char   u8;



#define VER         "0.1"



int gizmo_pwd(u8 *pwd);
u8 *find_string(u8 *data, int datasz, u8 *findme);
u8 *fdload(FILE *fd, int *fsize);
u8 *base64_decode(u8 *data, DWORD *size);
void std_err(void);
void wait_exit(int ret);



int main(int argc, char *argv[]) {
    FILE    *fd;
    int     buffsz;
    u8      *fname,
            *buff,
            *p,
            *l;

    fputs("\n"
        "Gizmo password decrypter "VER"\n"
        "by Luigi Auriemma\n"
        "e-mail: aluigi@autistici.org\n"
        "web:    aluigi.org\n"
        "\n", stdout);

    if(argc < 2) {
        printf("\n"
            "Usage: %s <AppSettings.xml/password>\n"
            "\n"
            "note: AppSettings.xml is located in %%APPDATA%%\\Gizmo5\n"
            "\n", argv[0]);
        wait_exit(1);
    }
    fname = argv[1];

    fd = fopen(fname, "rb");
    if(!fd) {
        p = strdup(fname);  // not needed
        gizmo_pwd(p);
        printf("\n  PASSWORD: %s\n", p);
    } else {
        buff = fdload(fd, &buffsz);
        fclose(fd);

        p = buff;
        while((p = find_string(p, buffsz - (p - buff), "<UserLongPass>"))) {
            gizmo_pwd(p);
            l = stristr(p, "</UserLongPass>");
            if(l) *l = 0;
            printf("\n  PASSWORD: %s\n", p);
            if(l) {
                p = l;
            } else {
                p++;
            }
        }

        free(buff);
    }

    wait_exit(0);
    return(0);
}



int gizmo_pwd(u8 *pwd) {
    HCRYPTPROV  hProv;
    HCRYPTHASH  hHash;
    HCRYPTKEY   hKey;
    DWORD   len;
    u8      key[] = "sip-47-giz-72-proj-2005-mr-43210",
            *tmp;

    len = strlen(pwd);
    tmp = base64_decode(pwd, &len);

    if(!CryptAcquireContext(
        &hProv,
        NULL,
        NULL,
        PROV_RSA_FULL,
        0)) return(-1);

    if(!CryptCreateHash(
        hProv,
        CALG_MD5,
        0,
        0,
        &hHash)) return(-1);

    if(!CryptHashData(
        hHash,
        key,
        strlen(key),
        0)) return(-1);

    if(!CryptDeriveKey(
        hProv,
        CALG_RC4,
        hHash,
        1,
        &hKey)) return(-1);

    if(!CryptDecrypt(
        hKey,
        0,
        TRUE,
        0,
        tmp,
        &len)) return(-1);

    tmp[len] = 0;

    strcpy(pwd, tmp);
    free(tmp);
    return(0);
}



u8 *find_string(u8 *data, int datasz, u8 *findme) {
    int     findmesz;
    u8      *p,
            *l;

    findmesz = strlen(findme);
    p = data;
    l = data + datasz - findmesz;
    for(p = data; p <= l; p++) {
        if(!strnicmp(p, findme, findmesz)) return(p + findmesz);
    }
    return(NULL);
}



u8 *fdload(FILE *fd, int *fsize) {
    struct stat xstat;
    int     size;
    u8      *buff;

    fstat(fileno(fd), &xstat);
    size = xstat.st_size;
    buff = malloc(size);
    if(!buff) std_err();
    fread(buff, 1, size, fd);
    *fsize = size;
    return(buff);
}



u8 *base64_decode(u8 *data, DWORD *size) {
    int     len,
            xlen,
            a   = 0,
            b   = 0,
            c,
            step;
    u8      *buff,
            *limit,
            *p;
    static const u8 base[128] = {   // supports also the Gamespy base64
        0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
        0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
        0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x3e,0x00,0x00,0x00,0x3f,
        0x34,0x35,0x36,0x37,0x38,0x39,0x3a,0x3b,0x3c,0x3d,0x00,0x00,0x00,0x00,0x00,0x00,
        0x00,0x00,0x01,0x02,0x03,0x04,0x05,0x06,0x07,0x08,0x09,0x0a,0x0b,0x0c,0x0d,0x0e,
        0x0f,0x10,0x11,0x12,0x13,0x14,0x15,0x16,0x17,0x18,0x19,0x3e,0x00,0x3f,0x00,0x00,
        0x00,0x1a,0x1b,0x1c,0x1d,0x1e,0x1f,0x20,0x21,0x22,0x23,0x24,0x25,0x26,0x27,0x28,
        0x29,0x2a,0x2b,0x2c,0x2d,0x2e,0x2f,0x30,0x31,0x32,0x33,0x00,0x00,0x00,0x00,0x00
    };

    if(!size || (*size < 0)) {      // use size -1 for auto size!
        len = strlen(data);
    } else {
        len = *size;
    }
    xlen = ((len >> 2) * 3) + 1;    // NULL included in output for text
    buff = malloc(xlen);
    if(!buff) return(NULL);

    p = buff;
    limit = data + len;

    for(step = 0; /* data < limit */; step++) {
        do {
            if(data >= limit) {
                c = 0;
                break;
            }
            c = *data;
            data++;
            if((c == '=') || (c == '_')) {  // supports also the Gamespy base64
                c = 0;
                break;
            }
        } while(c && ((c <= ' ') || (c > 0x7f)));
        if(!c) break;

        switch(step & 3) {
            case 0: {
                a = base[c];
                break;
            }
            case 1: {
                b = base[c];
                *p++ = (a << 2)        | (b >> 4);
                break;
            }
            case 2: {
                a = base[c];
                *p++ = ((b & 15) << 4) | (a >> 2);
                break;
            }
            case 3: {
                *p++ = ((a & 3) << 6)  | base[c];
                break;
            }
        }
    }
    *p = 0;

    len = p - buff;
    if(size) *size = len;
    if((len + 1) != xlen) {             // save memory
        buff = realloc(buff, len + 1);  // NULL included!
        if(!buff) return(NULL);
    }
    return(buff);
}



void std_err(void) {
    perror("\nError");
    wait_exit(1);
}



void wait_exit(int ret) {
    fprintf(stderr, "\npress RETURN to exit\n");
    fgetc(stdin);
    exit(ret);
}


